//Name: Justin Fiedler
//Project: HW3
//Course: CS 540

import java.io.*;
import java.util.Random;

public class HW3
{	public static boolean debug;
	private int inputCount;
	private int outputCount;
	private int curInput;
	private double[][] inputs;
	double[] layer1Outputs;
	double[] layer2Outputs;
	private String[] curOutput;
	private String[] outputs;
	private Neuron[] layer1;
	private Neuron[] layer2;

	public static void main(String args[])
	{	//java HW3 <names> <trainset> <testset> <hidden> <epochs> <alpha> <seed> [-d]
		debug = false;
		
		//String dir = "/afs/cs.wisc.edu/u/f/i/fiedler/";
		String names	= args[0];
		String train    = args[1];
		String test     = args[2];
		int hidden   	= Integer.parseInt(args[3]);
		int epochs   	= Integer.parseInt(args[4]);
		double alpha	= Double.parseDouble(args[5]);
		int seed	= Integer.parseInt(args[6]);

		
		if(args.length == 8  && args[7].equals("-d"))
		{    debug = true;
		}
		

		new HW3(names, train, test, hidden, epochs, alpha, seed);
	}

	public HW3(String names, String train, String test, int hidden, int epochs,
				double alpha, int seed)
	{	//read names file
		readNamesFile(names);
		//System.out.println("names done");

		//create neural net
		layer1 = new Neuron[hidden];
		layer2 = new Neuron[outputCount];

		setRandomWeights(hidden, seed);
		//System.out.println("weights set");

		//printNetwork();

	    //run train
		train(train, alpha, hidden, epochs);
		//System.out.println("train done");

	    //run test
		test(test, hidden);
		//System.out.println("test done");
	}

	//sets random weights to all nodes in the network
	public void setRandomWeights(int hidden, int seed)
	{	Random r = new Random(seed);
		double[] weights = new double[inputCount];
		//set hidden layer weights
		for(int j = 0; j < hidden; j++)
		{	for(int k = 0; k < inputCount; k++)
			{   //W(k,j) = r.nextDouble() - .5;
				weights[k] = r.nextDouble() - .5;
			}
			layer1[j] = new Neuron(weights);
		}

		weights = new double[hidden];
		//set output layer weights
		for(int i = 0; i < outputCount; i++)
		{	for(int j = 0; j < hidden; j++)
			{   //W(j,i) = r.nextDouble() - .5;
				weights[j] = r.nextDouble() - .5;
			}
			layer2[i] = new Neuron(weights);
		}
	}

	// runs training data through the network for 'epochs' epochs
	// and updates the weights once per epoch according to delta
	// values
	public void train(String train, double alpha, int hidden, int epochs)
	{	inputs = new double[100][inputCount];
		curOutput = new String[100];
		//read examples in from file
		try
		{	MyFileReader fr = new MyFileReader(train, ",");
			String line = fr.readLine();
			int i = 0;
			while(line != null)
			{	for(int j = 0; j < inputCount; j++)
				{	inputs[i][j] = Double.parseDouble(fr.nextToken());
				}
				curOutput[i] = fr.nextToken();
				i++;
				if(i >= inputs.length)
				{	inputs = resizeArray(inputs, 2 * inputs.length);
					curOutput = resizeArray(curOutput, 2 * curOutput.length);
				}

				line = fr.readLine();
			}
			inputs = resizeArray(inputs, i);
			curOutput = resizeArray(curOutput, i);
		}
		catch(FileNotFoundException ex)
		{	System.out.println(train + " not found");
			System.exit(0);
		}
		catch(IOException ex)
		{	System.out.println("Error reading from " + train);
			System.exit(0);
		}
		//System.out.println("File read");
		//cycle through data 'epochs' times
		for(int e = 0; e < epochs; e++)
		{	initialize_w_deltas();
			//propagate each example through the network
			for(curInput = 0; curInput < inputs.length; curInput++)
			{
				layer1Outputs = new double[hidden];
				layer2Outputs = new double[outputCount];

				run_current_inputs();

				//if(curInput == 1) printNetwork();

				calculate_deltas(curOutput[curInput]);

				//if(curInput == 1) printNetwork();

				calculate_w_deltas();

				/*
				if(curInput == 4)
				{	System.out.println(inputs[curInput][0]);
					System.out.println(inputs[curInput][1]);
					printNetwork();
				}
				*/
			}
			update_weights(alpha);
			//printNetwork();
			/*
			if(e%100 == 0)
			{	System.out.print("\nepoch: " + e);
			}
			else if(e%10 == 0)
			{	System.out.print('.');
			}
			*/
			
		}
		//System.out.println();
	}

	public void initialize_w_deltas()
	{	//System.out.println("initialize w deltas");
		//set d_w to zero for every weight
		for(int i = 0; i < layer1.length; i++)
		{	layer1[i].resetWDeltas();
		}
		for(int i = 0; i < layer2.length; i++)
		{	layer2[i].resetWDeltas();
		}
	}

	public void run_current_inputs()
	{	//System.out.println("run current inputs");
		//run inputs through first layer and save outputs
		for(int i = 0; i < layer1.length; i++)
		{	layer1Outputs[i] = layer1[i].output(inputs[curInput]);
		}

		//run layer1's outputs as inputs for layer2 and save outputs
		for(int i = 0; i < layer2.length; i++)
		{	layer2Outputs[i] = layer2[i].output(layer1Outputs);
		}
	}

	public void calculate_deltas(String curOutput)
	{	//System.out.println("calculate deltas");
		//calculate delta values for all nodes
		//generate correct output array
		double[] y = get_y_array(curOutput);

		//calcualte delta values for output layer (layer2)
		for(int i = 0; i < layer2.length; i++)
		{	layer2[i].setFinalLayerDelta(y[i]);
		}

		//calculate delta values for hidden layer (layer1)
		for(int i = 0; i < layer1.length; i++)
		{	double[] weights = new double[outputCount];
			double[] deltas  = new double[outputCount];
			for(int j = 0; j < outputCount; j++)
			{	weights[j] = (layer2[j].w())[i];
				deltas[j]  = layer2[j].delta();
			}
			layer1[i].setHiddenLayerDelta(weights, deltas);
		}
	}

	public void calculate_w_deltas()
	{	//System.out.println("caculate w deltas");
		//calculate delta wij for each weight, and add it to the old wij
		//calculate delta w's for layer1
		for(int i = 0; i < layer1.length; i++)
		{	layer1[i].setWDeltas(inputs[curInput]);
		}

		//calculate delta w's for layer2
		for(int i = 0; i < layer2.length; i++)
		{	layer2[i].setWDeltas(layer1Outputs);
		}
	}

	public void update_weights(double alpha)
	{	//System.out.println("update weights");
		//update weights for each node
		//update layer1's weights
		for(int i = 0; i < layer1.length; i++)
		{	layer1[i].update_weights(alpha);
		}
		//update layer2's weights
		for(int i = 0; i < layer2.length; i++)
		{	layer2[i].update_weights(alpha);
		}
	}

	// runs test data through the network and compares 'best'
	// approximation to actual catergory
	public void test(String test, int hidden)
	{	int correct = 0;
		int num_examples = 0;

		inputs = new double[1][inputCount];
		curInput = 0;

  		try
		{	MyFileReader fr = new MyFileReader(test, ",");
			String line = fr.readLine();
			while(line != null)
			{	for(int i = 0; i < inputCount; i++)
				{	inputs[0][i] = Double.parseDouble(fr.nextToken());
				}
				String curOutput = fr.nextToken();

				layer1Outputs = new double[hidden];
				layer2Outputs = new double[outputCount];

				run_current_inputs();

				int max = 0;
				for(int i = 1; i < layer2Outputs.length; i++)
				{	max = layer2Outputs[i] > layer2Outputs[max] ? i : max;
				}

				double[] y = get_y_array(curOutput);
				if(y[max] == 1) correct++;

				if(debug == true) System.out.println(outputs[max]);

				num_examples++;

				line = fr.readLine();
			}
		}
		catch(FileNotFoundException ex)
		{	System.out.println(test + " not found");
			System.exit(0);
		}
		catch(IOException ex)
		{	System.out.println("Error reading from " + test);
			System.exit(0);
		}

		System.out.println(((double)correct)/num_examples);
	}

    /* readNamesFile
	 * accepts the name of a file from which to read the number of inputs to
	 * expect from each example and the possible output catergories
     */
	public void readNamesFile(String fileName)
	{	outputs = new String[10];
		try
		{	MyFileReader fr = new MyFileReader(fileName, ",");
			String line = fr.readLine();
			inputCount = Integer.parseInt(fr.nextToken());
			line = fr.readLine();
			int i = 0;
			for(i = 0; fr.hasMoreTokens(); i++)
			{	if(i >= outputs.length)
				{	outputs = resizeArray(outputs, 2 * outputs.length);
				}
				outputs[i] = fr.nextToken();
			}
			outputs = resizeArray(outputs, i);
			outputCount = outputs.length;
		}
		catch(FileNotFoundException e)
		{	System.out.println(fileName + " not found");
			System.exit(0);
		}
		catch(IOException e)
		{	System.out.println("Error reading from " + fileName);
			System.exit(0);
		}
	}

	// returns the appropiate double array representing the
	// given output
	private double[] get_y_array(String output)
	{	double[] y = new double[outputs.length];
		for(int i = 0; i < outputs.length; i++)
		{	y[i] = 0;
			if(outputs[i].equals(output)) y[i] = 1;
		}
		return y;
	}

	/* resizeArray
	 * returns a String array of size 'newSize' containing the
	 * contents of 'arr'
	 */
	private String[] resizeArray(String[] arr, int newSize)
	{	String[] temp = new String[newSize];
		for(int i = 0; (i < arr.length && i < newSize); i++)
		{	temp[i] = arr[i];
		}
		return temp;
	}

	/* resizeArray
	 * returns a double[][] array of size 'newSize' containing the
	 * contents of 'arr'
	 */
	private double[][] resizeArray(double[][] arr, int newSize)
	{	double[][] temp = new double[newSize][inputCount];
		for(int i = 0; (i < arr.length && i < newSize); i++)
		{	temp[i] = arr[i];
		}
		return temp;
	}

	//prints out all neurons in the network
	public void printNetwork()
	{	System.out.println("**HIDDEN LAYER**");
		for(int i = 0; i < layer1.length; i++)
		{	System.out.println("Node "+ i + " " + layer1[i]);
		}

		System.out.println("**OUTPUT LAYER**");
		for(int i = 0; i < layer2.length; i++)
		{	System.out.println("Node "+ i + " " + layer2[i]);
		}
	}
}
